import logging

# Configure logging
logging.basicConfig(level=logging.INFO)  # Set level to INFO or higher
from psycopg2 import sql, connect
from psycopg2.extras import DictCursor,RealDictCursor
from abstract_utilities import make_list, SingletonMeta, is_number
from abstract_security import get_env_value
import psycopg2,time
env_path = '/home/solcatcher/.env'

def get_env_val(key):
    return get_env_value(key=key, path=env_path)

RABBITMQ_HOST = get_env_val("SOLCATCHER_AMQP_HOST") or 'solcatcher.io'
RABBITMQ_PORT = get_env_val("SOLCATCHER_AMQP_PORT") or '5672'
RABBITMQ_USER = get_env_val("SOLCATCHER_AMQP_USER") or 'solcatcher'
RABBITMQ_NAME = get_env_val("SOLCATCHER_AMQP_NAME") or 'solcatcher'
RABBITMQ_PASSWORD = get_env_val("SOLCATCHER_AMQP_PASSWORD") or 'solcatcher123'

POSTGRESQL_HOST = get_env_val("SOLCATCHER_POSTGRESQL_HOST") or 'solcatcher.io'
POSTGRESQL_PORT = get_env_val("SOLCATCHER_POSTGRESQL_PORT") or '5432'
POSTGRESQL_USER = get_env_val("SOLCATCHER_POSTGRESQL_USER") or 'solcatcher'
POSTGRESQL_NAME = get_env_val("SOLCATCHER_POSTGRESQL_NAME") or 'solcatcher'
POSTGRESQL_PASSWORD = get_env_val("SOLCATCHER_POSTGRESQL_PASSWORD") or 'solcatcher123!!!456'

DB_URL = f"postgresql://{POSTGRESQL_USER}:{POSTGRESQL_PASSWORD}@{POSTGRESQL_HOST}:{POSTGRESQL_PORT}/{POSTGRESQL_NAME}"

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

def get_connection():
    """Establish a database connection."""
    return psycopg2.connect(
        dbname=POSTGRESQL_NAME,
        user=POSTGRESQL_USER,
        password=POSTGRESQL_PASSWORD,
        host=POSTGRESQL_HOST,
        port=POSTGRESQL_PORT
    )
def query_data_as_dict(query, values=None, error="Error executing query:"):
    """Execute a query and return a list of dictionaries using RealDictCursor."""
    with get_connection() as conn:
        with conn.cursor(cursor_factory=RealDictCursor) as cursor:
            try:
                cursor.execute(query, values)
                result = cursor.fetchall()
                # Return the result as a list of dicts
                return result
            except Exception as e:
                conn.rollback()
                logging.error("%s %s\nValues: %s\n%s", error, query, values, e)
                return []

def get_all_table_names(schema='public'):
    """Fetch all table names from a specified schema."""
    conn = get_connection()
    query = """
        SELECT table_name
        FROM information_schema.tables
        WHERE table_schema = %s AND table_type = 'BASE TABLE';
    """
    try:
        with conn.cursor() as cursor:
            cursor.execute(query, (schema,))
            table_names = [row[0] for row in cursor.fetchall()]
            return table_names
    except Exception as e:
        print(f"An error occurred in get all tyables: {e}")
    finally:
        conn.close()


def get_query_result(query, conn):
    """Executes a query and returns the results or commits the transaction."""
    with conn.cursor() as cursor:
        cursor.execute(query)
        if query.strip().lower().startswith("select"):
            return cursor.fetchall()  # Return data for SELECT queries
        conn.commit()

def query_data(query, values=None, error="Error executing query:", zipRows=True):
        """Execute a query and handle transactions with error management."""
        logging.info(f"query = {query} and values = {values}")
        with get_connection() as conn:
            # Choose the cursor type based on whether you want to zip rows with column names
            cursor_factory = DictCursor if zipRows else None
            with conn.cursor(cursor_factory=cursor_factory) as cursor:
                try:
                    cursor.execute(query, values)
                    result = cursor.fetchall()
                    # Log the first row to see its structure
                    if result:
                        logging.info("First row data structure: %s", result[0])
                    return result
                except Exception as e:
                    conn.rollback()
                    logging.error("%s %s\nValues: %s\n%s", error, query, values, e)
def get_query_result(query, conn=None):
    conn = conn or getConnection()
    """Executes a query and returns the results or commits the transaction."""
    with conn.cursor() as cursor:
        cursor.execute(query)
        if query.strip().lower().startswith("select"):
            return cursor.fetchall()  # Return data for SELECT queries
        conn.commit()
def aggregate_rows(query, values=None, errorMsg='Error Fetching Rows'):
    conn = get_connection()
    cursor = None
    try:
        cursor = conn.cursor(cursor_factory=DictCursor)
        cursor.execute(query, values)

        # Handle SELECT queries differently
        if query.strip().upper().startswith("SELECT"):
            try:
                results = cursor.fetchall()
                return results  # Return fetched rows
            except psycopg2.ProgrammingError as e:
                # No rows to fetch—return an empty list
                logger.warning(f"No rows to fetch for query: {query}")
                return []

        # Commit and return empty for non-SELECT queries
        conn.commit()
        return []
    except Exception as e:
        error_details = f"{errorMsg}:\n{str(e)}\nTraceback:\n{traceback.format_exc()}"
        logger.error(error_details)
        return []
    finally:
        if cursor:
            cursor.close()



from .query_functions.queries import aggregate_rows,query_data,query_data_as_dict
from .query_functions.fetch_utils import fetch_any_combo,get_db_from,getZipRows
from ..utils import derive_timestamp,if_list_get_single,get_signature,get_sorted_txn_history



from typing import *
from ...utils import *
from abstract_utilities import make_list,SingletonMeta
class rowManager(metaclass=SingletonMeta):
    def __init__(self, 
                 txn_columns: Optional[List[str]] = None, 
                 main_columns: Optional[List[str]] = None, 
                 general_retain: Optional[List[str]] = None,
                 pools: Optional[List[str]] = None, 
                 pool_rows: Optional[List[str]] = None, 
                 pool_values: Optional[List[str]] = None,
                 pool_ids: Optional[List[str]] = None):
        if not hasattr(self, 'initialized'):
            self.reset()
    def reset(self):
        self.initialized=True
        self.txn_columns = txn_columns or []
        self.main_columns = main_columns or []
        self.general_retain = general_retain or []
        self.pools = pools or []
        self.pool_rows = pool_rows or []
        self.pool_values = pool_values or []
        self.pool_ids = pool_ids or []
    def add_to_pools(self,pools = None,pool_rows = None,pool_data=None,pool_values = None,pool_ids = None):
        for pool in pools,pool_rows,pool_values,pool_ids:
            if pools:
                self.pools = pools
            if pool_rows:
                self.pool_rows = pool_rows
            if pool_values:
                self.pool_values = pool_values
            if pool_ids:
                self.pool_ids = pool_ids
            if pool_data:
                self.pool_data = pool_data

def sift_key_values(dict_obj: Dict[str, Any], keys: List[str], out: Optional[Dict[str, Any]] = None) -> Tuple[Dict[str, Any], List[str]]:
    """
    Attempt to extract the specified keys from dict_obj into the dictionary 'out'.
    Returns a tuple of (out, remaining_keys) where remaining_keys are those not found.
    """
    if out is None:
        out = {}
    remaining_keys = []
    dict_obj = if_list_get_single(dict_obj)

    for key in keys:
        if key not in out:
            out[key]=None
        value = dict_obj.get(key)
        if not value and key == 'timestamp':
            value = get_timestamps(dict_obj)
        if value:
            out[key] = value
        else:
            remaining_keys.append(key)
    return out, remaining_keys


def get_main_data_columns_from_row(row: Dict[str, Any], keys: List[str]) -> Dict[str, Any]:
    """
    Extracts main data columns from a single row by first checking the row itself
    and then (if needed) checking nested sub-dictionaries such as 'meta_data' and 'tcns'.
    
    Args:
        row: The input row as a dictionary.
        keys: A list of keys to look for (a copy of the master key list).
    
    Returns:
        A dictionary containing the found key-value pairs.
    """
    # Work on a copy of keys to avoid modifying the original list.
    remaining_keys = keys.copy()
    result: Dict[str, Any] = {}

    # First pass: extract from the main row dictionary.
    result, remaining_keys = sift_key_values(row, remaining_keys, result)
    if remaining_keys:
        # Second pass: if not all keys were found, try sub-dictionaries.
        for sub_field in ['meta_data', 'tcns']:
            if not remaining_keys:
                break  # All keys found; no need to continue.
            sub_objs = row.get(sub_field)
            for sub_obj in make_list(sub_objs):
                if sub_obj and isinstance(sub_obj, dict):
                    result, remaining_keys = sift_key_values(sub_obj, remaining_keys, result)
            # Optionally, if sub_obj might be a list of dictionaries, you could iterate over them.
    
    return result


def get_main_data_columns(rows: Any, main_columns: List[str]) -> List[Dict[str, Any]]:
    """
    Processes a collection of rows to extract main data columns based on the provided main_columns list.
    
    Args:
        rows: A single row or a list of rows.
        main_columns: The list of columns (keys) to extract.
    
    Returns:
        A list of dictionaries, each representing the extracted key-value pairs for a row.
    """
    pools: List[Dict[str, Any]] = []
    pool_rows: List[Dict[str, Any]] = []
    pool_values: List[Dict[str, Any]] = []
    pool_ids: List[Dict[str, Any]] = []
    # Optionally, if you need to use ColumnsManager for more complex logic later,
    # you could initialize it here.
    columns_mgr = rowManager(general_retain=main_columns)
    
    for row in make_list(rows):
        row = dict(row)
        # Always work with a copy of the keys to avoid modifying the original list.
        pools.append(row)
        pool_id = get_main_data_columns_from_row(row,['pool_id'])
        pool_ids.append(pool_id.get('pool_id'))
        pool_data = get_main_data_columns_from_row(row, columns_mgr.general_retain.copy())
        pool_values.append(list(pool_data.values()))
        pool_rows.append(pool_data)
    columns_mgr.add_to_pools(pools = pools,pool_rows = pool_rows,pool_values = pool_values,pool_ids = pool_ids)
    return pools,pool_rows,pool_data,pool_values,pool_ids



import asyncio
import json
from datetime import datetime  # Import the datetime class directly.
import requests  # Consider replacing with aiohttp for async requests

from abstract_solcatcher import *
from abstract_utilities import *
from abstract_security import get_env_value
from abstract_apis import postRpcRequest, asyncPostRpcRequest, get_headers
from .db_funcs import get_pair, query_data

logger = get_logFile('pair_data')
PUMP_FUN_PROGRAMID = '6EF8rrecthR5Dkzon8Nwu78hRvfCKubJ14M5uBEwF6P'

fallback_url = 'https://rpc.ankr.com/solana/c3b7fd92e298d5682b6ef095eaa4e92160989a713f5ee9ac2693b4da8ff5a370'

# Custom JSON encoder to handle datetime objects
class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)

# Custom dumps function using the above encoder
def custom_json_dumps(data):
    return json.dumps(data, cls=DateTimeEncoder)

def serialize_datetimes(obj):
    """
    Recursively converts datetime objects in a dict or list into ISO-formatted strings.
    """
    if isinstance(obj, dict):
        return {k: serialize_datetimes(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [serialize_datetimes(item) for item in obj]
    elif isinstance(obj, datetime):
        return obj.isoformat()
    return obj

async def call_solcatcher_db(endpoint, **kwargs):
    # NOTE: Using requests.post in an async function is blocking.
    # Consider using aiohttp instead.
    response = requests.post(
        url=f'https://solcatcher.io/dbCalls/{endpoint}',
        data=json.dumps(kwargs),
        headers=get_headers()
    )
    response = get_response(response)
    if isinstance(response, dict):
        response = response.get('result', response)
    return response

async def fetch_transaction(signature):
    method = 'getTransaction'
    params = [signature, {"maxSupportedTransactionVersion": 0}]
    response = await asyncPostRpcRequest(
        url=fallback_url,
        method=method,
        params=params,
        response_result='result'
    )
    return response

async def fetch_signatures_from_address(address):
    method = 'getSignaturesForAddress'
    params = [address]
    response = await asyncPostRpcRequest(
        url=fallback_url,
        method=method,
        params=params,
        response_result='result'
    )
    return response

async def get_parsed_log_data(transaction):
    txn_data = {
        'signatures': get_any_value(transaction, 'signatures'),
        'slot': get_any_value(transaction, 'slot'),
        'signature': get_any_value(transaction, 'signatures')[0],
        'program_id': PUMP_FUN_PROGRAMID,
        'logs': get_any_value(transaction, 'logMessages')
    }
    return await async_call_solcatcher_ts('full-process-logs', **txn_data)

async def get_creation_pubkey(mint):
    meta_calls = ['get-or-fetch-metadata', 'fetch-meta']
    publickeys = None
    for call in meta_calls:
        metaData = await call_solcatcher_ts(call, mint=mint, url=fallback_url)
        if isinstance(metaData, int):
            metaData = await call_solcatcher_ts('get-or-fetch-metadata', mint=mint, url=fallback_url)
        publickeys = get_any_value(metaData, 'publicKey')
        if publickeys:
            break
    publickeys = [key for key in publickeys if key != mint] if publickeys else []
    return publickeys

async def decode_instruction_data(data):
    return await async_call_solcatcher_ts(endpoint='decode-instruction-data', data=data)

async def get_signatures_from_transaction(transaction):
    signatures = get_any_value(transaction, 'signatures')
    return signatures

async def get_creation_signature_from_mint(mint):
    addresses = await get_creation_pubkey(mint)
    if addresses:
        meta_address = addresses[0]
        sigs = await fetch_signatures_from_address(meta_address)
        signatures = [sig.get('signature') for sig in sigs]
        return signatures
    return None
async def get_transaction_from_signatures(signatures):
    transaction = await fetch_transaction(signature)

async def get_creation_transaction_from_mint(mint):
    signatures = await get_creation_signature_from_mint(mint)
    if signatures:
        signature = signatures[-1]
        transaction = await fetch_transaction(signature)
        return transaction,signature
    return None
async def get_creation_parsed_logs_from_mint(mint):
    transaction = await get_creation_transaction_from_mint(mint)
    logger.info(f"transaction = {transaction}")
    parsed_logs = await get_parsed_log_data(transaction)
    signatures = await get_signatures_from_transaction(transaction)
    for signature in signatures:
        log_data = await fetch_log_data(signature)
        if log_data:
            return signature, log_data
    return None, None

async def get_user_address(parsed_logs):
    if parsed_logs:
        logs = parsed_logs.get('logs', [])
        datas = [entry.get('data')[0] for entry in logs if entry.get('data') and 'Instruction: Create' in entry.get('logs', '')]
        for data in datas:
            if 'Instruction: Create' in data.get('logs', ''):
                raw_data = data.get('data')
                decoded_data = await decode_instruction_data(data=raw_data)
                user_address = decoded_data.get('user_address')
                return user_address
    return None
async def get_data_from_parsed_logs(parsed_logs):
    datas = []
    for entry in parsed_logs:
        if entry.get('data') and 'Instruction: Create' in entry.get('logs', ''):
            decoded = await decode_instruction_data(data=entry.get('data')[0])
            datas.append(decoded)
    return get_if_single(datas)
async def get_creation_instruction_data(mint):
    transaction = await get_creation_transaction_from_mint(mint=mint)
    parsed_logs = await get_parsed_log_data(transaction)
    return await get_data_from_parsed_logs(parsed_logs)
async def get_sanitized_pair(mint):
    pair = get_pair(mint=mint)
    while True:
        if pair and isinstance(pair,list) and len(pair) == 1:
            pair = pair[0]
        else:
            break
    pair = serialize_datetimes(pair)
    return pair
async def send_to_txn_queue(mint):
    pairMsg = {}
    transaction,signature = await get_creation_transaction_from_mint(mint=mint)
    pairMsg['signature']=signature
    pairMsg['logs'] = get_any_value(transaction,'logMessages')
    pairMsg['program_id'] = PUMP_FUN_PROGRAMID
    pairMsg['slot'] = get_any_value(transaction,'slot')
    await async_call_solcatcher_ts('sendTo-logIntake',**pairMsg);
    return signature
async def async_update_pair(mint):
    pair = await get_sanitized_pair(mint)
    transaction,signature = await get_creation_transaction_from_mint(mint=mint)
    pair['signature']=signature
    parsed_logs = await get_parsed_log_data(transaction)
    pair['log_id']=parsed_logs.get('log_id')
    decoded_data = await get_data_from_parsed_logs(parsed_logs.get('parsedLogs'))
    pair['user_address']=decoded_data.get('user_address')
    result = await async_call_solcatcher_ts('update-pair', pairData=pair)
    return pair
def update_pair(mint):
    return asyncio.run(async_update_pair(mint))




from ...utils import *
def display_user_data(user_txns, user, txn_columns,window):
    """
    Display individual user's transaction data in the GUI.

    Parameters:
        user_data (dict): Data for the selected user.
        user (str): User address.
        window (sg.Window): The PySimpleGUI window.
    """
    #if user not in user_data:
    #    sg.popup_error(f"No data found for user: {user}")
    #    return
    txn_table_data = [[
        get_timestamps(txn),
        float(get_txn_price(txn) or 0),
        float(get_sol_amount(txn) or 0),
        float(get_token_amount(txn) or 0),
        "Buy" if get_is_buy(txn) else "Sell",
        get_signature(txn)
    ] for txn in user_txns]
    # Update table
    #window['-USER_TXN_TABLE-'].update(values=txn_table_data)
    window['-USER_TXN_TABLE-'].update(
                    values=[[function(txn) for function in [get_timestamps, get_txn_price, get_sol_amount, get_token_amount, get_is_buy, get_signature]] for txn in user_txns]
                )
    # Generate chart
    #chart_image_path = f"{user}_chart.png"

    #generate_price_chart(txn_table_data, save_path=chart_image_path)

    # Update chart display
    #window['-USER_CHART-'].update(filename=chart_image_path)

    # Update user-specific stats
    #user_stats = user_data[user]
    #window['-USER_STATS-'].update(
    #    f"User: {user}\n"
    #    f"Profit (SOL): {user_stats['profits']['sol']}\n"
    #    f"Average Price: {user_stats['avgPrice']['avg']}\n"
    #    f"Volume (Buy): {user_stats['volume']['buy']}\n"
    #    f"Volume (Sell): {user_stats['volume']['sell']}\n"
    #    f"Total Volume: {user_stats['volume']['total']}"
    #)
def set_user_info(txn_history,window):
    user_data = tally_profits(txn_history)
    user_wallets = list(user_data.keys())
    window['-USER_COMBO-'].update(values=user_wallets,value=user_wallets[0])
    return user_data
def tally_users(user_data,window,values):
    user_wallet = values['-USER_COMBO-'][0]
    display_user_data(user_data, user_wallet, window)
    #generate_price_chart(txn_history=user_txns['txns'],save_path="user_chart.png")
    #window['-USER_CHART-'].update(filename="user_chart.png")



from ..queries import query_data_as_dict,query_data,derive_timestamp
def fetch_filtered_transactions_paginated(
    sol_amount, 
    operator=">",
    timestamp_operator=None, 
    limit=200, 
    offset=0,
    years=0,
    months=0,
    weeks=0,
    days=0,
    hours=0,
    minutes=0,
    seconds=0,
    timestamp=None,
    *args,
    **kwargs
):
    query = f"""
        SELECT
            p.*, 
            m.*,
            t.*
        FROM 
            transactions t
        JOIN 
            pairs p ON t.pair_id = p.id
        LEFT JOIN 
            metadata m ON p.meta_id = m.id
        WHERE 
            t.signature IN (
                SELECT signature FROM pairs WHERE signature IS NOT NULL
            )
        AND 
            t.program_id = p.program_id
        AND 
            EXISTS (
                SELECT 1
                FROM jsonb_array_elements(t.tcns) AS elem
                WHERE (elem ->> 'sol_amount')::numeric {operator} %s
    """
    params = [sol_amount]

    ts = derive_timestamp(
            years=years,
            months=months,
            weeks=weeks,
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            timestamp=timestamp)
    
    if ts and timestamp_operator:
        query += f" AND (elem ->> 'timestamp')::bigint {timestamp_operator} %s"
        params.append(int(ts))
    
    query += ") ORDER BY t.updated_at DESC LIMIT %s OFFSET %s;"
    params.extend([limit, offset])
    
    return query_data_as_dict(query, params)




from .queries import get_db_from,fetch_any_combo
from .call_log_data import get_signature_from_log_id
def get_transaction_from(columnNames='*',searchColumn=None,searchValue=None,count=False):
    return get_db_from('transactions',columnNames=columnNames,searchColumn=searchColumn,searchValue=searchValue)
def count_transactions_for_pair(pair_id):
    return get_transaction_from(searchColumn="pair_id",searchValue=pair_id,count=True)
def get_transaction_from_signature(signature):
    return get_transaction_from(searchColumn="signature",searchValue=signature)
def get_transaction_from_txn_id(txn_id):
    return get_transaction_from(searchColumn="id",searchValue=txn_id)
def get_transaction_from_log_id(log_id):
    signature = get_signature_from_log_id(log_id)
    return get_transaction_from_signature(signature)
def get_all_txns_from_pair_id(pair_id):
    return get_transaction_from(searchColumn="pair_id",searchValue=pair_id)
def get_transactions(txn_id=None,log_id=None,signature=None,pair_id=None):
    if pair_id:
        return get_all_txns_from_pair_id(pair_id)
    if txn_id:
        return get_transaction_from_txn_id(txn_id)
    if log_id:
        return get_transaction_from_log_id(log_id)
    if signature:
        return get_transaction_from_signature(signature)




from ...utils.utils import *
from datetime import datetime
from ...query_functions import get_any_pair,fetch_any_combo,get_transactions,get_pairs_from_user_wallets
#from ..queries import fetch_any_combo,get_transactions,get_pairs_from_user_wallets
from ...managers.meta_data_manager import metaDataManager
from .pool_metrics import *
from ..user_txns.tally_profits import tally_profits
metaDataMgr = metaDataManager()

def rank_supply(supply, initial_pool_tokens):
    if supply is None or supply == 0 or initial_pool_tokens == None or initial_pool_tokens == 0:
        return 0
    percent_tokens_available = initial_pool_tokens / supply
    deviation = 1 - percent_tokens_available
    score = 20 - abs(10 - deviation)
    if score < 0:
        score = 0
    return score

def metaDataRanking(metaDataVars):
    total_score = 0
    metaDataScores = {
        'image':  ['bools', 2, 0],
        'uri':    ['bools', 2, 0],
        'symbol': ['bools', 2, 0],
        'name':   ['bools', 2, 0],
        'mintAuthority':   ['processed', 0, 2],
        'freezeAuthority': ['processed', 0, 2],
        'twitter': ['bools', 6, 0],
        'website': ['bools', 6, 0],
        'isMutable': ['processed', 2, 0],
        'primarySaleHappened': ['processed', 6, 0],
        'description': ['bools', 6, 0],
    }
    for key, values in metaDataScores.items():
        # Ensure we safely access the metaDataVars keys
        if metaDataVars.get(values[0]):
            total_score += values[1]
        else:
            total_score += values[-1]
    return total_score

def if_signature_get_txn(obj):
    if not isinstance(obj, str):
        return obj
    if len(obj) == 88:
        obj = fetch_any_combo(columnNames='*', tableName='transactions', searchColumn='signature', searchValue=obj)
        obj = make_list(obj)
    return obj

def rank_initial_liquidity(initial_sol_amount):
    if initial_sol_amount >= 20:
        return 100
    if initial_sol_amount >= 10:
        remainder = 20 - initial_sol_amount
        return 100 - (remainder * 3)
    if initial_sol_amount >= 5:
        remainder = 10 - initial_sol_amount
        return 70 - (remainder * 4)
    if initial_sol_amount < 5:
        remainder = 5 - initial_sol_amount
        return 30 - (remainder * 6)
    return 0

def score_total_profits_enhanced(total_profits, num_txns, creation_time, last_significant_txn_time,
     norm_params, weights, decay_rate=0.1, current_time=None):
    """
    Enhanced scoring function with time decay applied to last significant transaction.
    """
    if current_time is None:
        current_time = datetime.utcnow()

    # Convert timestamps to datetime objects
    creation_dt = datetime.strptime(creation_time, '%Y-%m-%dT%H:%M:%SZ')
    last_significant_dt = datetime.strptime(last_significant_txn_time, '%Y-%m-%dT%H:%M:%SZ')

    # Calculate time differences in days
    time_since_creation = (current_time - creation_dt).total_seconds() / 86400
    time_since_last_significant_txn = (current_time - last_significant_dt).total_seconds() / 86400

    # Apply time decay (assuming time_decay is defined in pool_metrics)
    decay_creation = time_decay(time_since_creation, decay_rate)
    decay_last_significant = time_decay(time_since_last_significant_txn, decay_rate)

    # Normalize metrics (assuming min_max_normalize is defined in pool_metrics)
    V_norm = min_max_normalize(total_profits, norm_params['total_profits_min'], norm_params['total_profits_max'])
    N_norm = min_max_normalize(num_txns, norm_params['num_txns_min'], norm_params['num_txns_max'])
    C_norm = min_max_normalize(decay_creation, norm_params['creation_time_min'], norm_params['creation_time_max'])
    L_norm = min_max_normalize(decay_last_significant, norm_params['last_significant_txn_time_min'], norm_params['last_significant_txn_time_max'])

    # Calculate weighted sum
    activity_score = (
        weights['V'] * V_norm +
        weights['N'] * N_norm +
        weights['C'] * C_norm +
        weights['L'] * L_norm
    )
    return activity_score
def get_genesis_txn(signature):
    genesisTxn = get_transactions(signature=signature)
    genesisTxn = if_list_get_single(genesisTxn)
    genesisTxn =flatten_txn(genesisTxn)
    return genesisTxn
class TransactionProcessor:
    def __init__(self, pair):
        """Initialize the transaction processor with a pair."""
        pair = get_any_pair(pair)
        self.pair = pair
        if 'error' not in self.pair:
            self.mint = self.get_mint()
            self.pair_id = self.pair.get('id')
            self.owner_address = self.pair.get('user_address')
            self.genesis_signature = self.pair.get('signature')
            self.genesisTxn = get_genesis_txn(self.genesis_signature)
            self.creation_time = get_timestamps(self.genesisTxn)
            self.initial_sol_amt = get_sol_amount(self.genesisTxn) or 0
            self.initial_token_amt = get_token_amount(self.genesisTxn) or 0
            
            self.all_txns = [flatten_txn(txn) for txn in get_transactions(pair_id=self.pair_id)]
            self.all_txns = [txn for txn in self.all_txns if get_signature(txn) != self.genesis_signature]
            self.user_wallets = self.get_all_user_wallets(self.all_txns)
            if 'error' not in self.all_txns:
                self.total_profits = self.tally_profits(self.all_txns)
                self.metaDataVars = metaDataMgr.processMetaData(self.mint, imageData=False)
                self.metaDataBools = self.metaDataVars.get('bools')
                self.assigned_account = []#get_assigned_account(self.owner_address)
                self.pool_metrics = calculateMetrics([{**self.pair, 'transactions': self.all_txns}])
                self.profits = self.tally_profits(self.all_txns)
                self.total_sol_volume = self.get_total_sol_volume(self.all_txns)
                self.total_profits = self.total_profit(self.profits)
                #self.previous_pairs = get_pairs_from_user_wallets(self.user_wallets)
                self.scores = self.ranking()
             
    def ranking(self):
        # Use a supply value from metaDataMgr; assuming datas_js exists and has a 'supply' key
        supply = metaDataMgr.datas_js.get('supply', 0)
        supply_score = rank_supply(supply, self.initial_token_amt)  # target 90% in pool initial; 20 pts
        metadata_score = metaDataRanking(self.metaDataVars)  # target all metas in good standing; 36 pts
        initial_liquidity_score = rank_initial_liquidity(self.initial_sol_amt)  # target above 20 initial sol; 100 pts            
        # Calculate active_time and dead_time based on pool_metrics and current time
        last_sig_time = get_time_of_last_significant_txn(self.pool_metrics) if self.pool_metrics else None
        if last_sig_time and self.creation_time and isinstance(last_sig_time,int or float):
            try:
                active_time = last_sig_time - self.creation_time
                dead_time = time.time() - last_sig_time if last_sig_time else None
            except:
                active_time = None
                dead_time=None
        else:
            active_time = None
            dead_time=None
        #previous_pairs_grades = self.rank_previous_pairs(self.previous_pairs)  # 150 pts total
        #previous_pair_grade = sum(previous_pairs_grades) / len(previous_pairs_grades) if previous_pairs_grades else 0
        total_score = supply_score + metadata_score + initial_liquidity_score
        pair_grade = total_score / (20 + 36 + 100)
        #previous_pairs_grades.append(pair_grade)
        #overall_grade = sum(previous_pairs_grades) / len(previous_pairs_grades)
        scores = {
            "active_time": active_time,
            "last_sig_time": last_sig_time,
            "dead_time": dead_time,
            "total_volume": self.total_sol_volume,
            "active_time": active_time,
            "supply_score": supply_score,
            "metadata_score": metadata_score,
            "initial_sol_score": initial_liquidity_score,
            #"previous_pair_grade": previous_pair_grade,
            "pair_grade": pair_grade,
            #"overall_grade": overall_grade
        }
        return scores

    def rank_previous_pairs(self, pairs):
        total_grades = []
        for pair in pairs:
            try:
                pair_id = pair.get('id')
                genesis_signature = pair.get('signature')
                mint = pair.get('mint')
                genesisTxn = get_transactions(signature=genesis_signature)
                genesisTxn = if_list_get_single(genesisTxn)
                genesisTxn =flatten_txn(genesisTxn)
                pair_txns = [flatten_txn(txn) for txn in get_all_txns_for_pair_id(pair_id=pair_id) if get_signature(txn) != genesis_signature]
                creation_time = get_timestamps(genesisTxn) or 0
                initial_sol_amt = get_sol_amount(genesisTxn) or 0
                initial_token_amt = get_token_amount(genesisTxn) or 0
                profits = self.tally_profits(pair_txns)
                total_profits = self.total_profit(profits)
                time_alive = self.calculate_time_interval(pair_txns)
                pool_metrics = calculateMetrics([{**pair, 'transactions': pair_txns}])
                metaDataVars = metaDataMgr.processMetaData(mint, imageData=False)
                metadata_score = metaDataRanking(metaDataVars)
                metaDataBools = metaDataVars.get('bools')
                metaDataProcessed = metaDataVars.get('processed', {})
                supply_score = rank_supply(get_supply(metaDataProcessed) or 0, initial_token_amt)
                initial_liquidity_score = rank_initial_liquidity(initial_sol_amt)
                total_score = supply_score + metadata_score + initial_liquidity_score
                grade = total_score / (20 + 36 + 100)
                total_grades.append(grade)
            except:
                pass
        return total_grades

    def get_total_profits(self, txns=None):
        if txns:
            return self.tally_profits(txns)
        if hasattr(self, 'total_profits'):
            return self.total_profits
        self.total_profits = self.tally_profits()
        return self.total_profits

    def get_all_user_wallets(self, txns=None):
        if txns:
            total_profits = self.get_total_profits(txns)
            return list(total_profits.keys())
        if hasattr(self, 'user_wallets'):
            return self.user_wallets
        total_profits = self.get_total_profits()
        self.user_wallets = list(total_profits.keys())
        return self.user_wallets

    def get_mint(self, pair=None):
        if pair:
            return pair.get('mint')
        if hasattr(self, 'mint'):
            return self.mint
        self.mint = self.pair.get('mint')
        return self.mint

    def get_genesis_txn(self, pair=None):
        if pair:
            genesisTxn, _ = self.separate_genesis_txn(pair)
            return genesisTxn
        if hasattr(self, 'genesisTxn'):
            return self.genesisTxn
        self.genesisTxn, self.all_txns = self.separate_genesis_txn()
        return self.genesisTxn

    def get_all_txns(self, pair=None):
        if pair:
            _, all_txns = self.separate_genesis_txn(pair)
            return all_txns
        if hasattr(self, 'all_txns'):
            return self.all_txns
        self.genesisTxn, self.all_txns = self.separate_genesis_txn()
        return self.all_txns

    def get_initial_sol_amt(self, pair=None):
        genesisTxn = self.get_genesis_txn(pair)
        if pair:
            return self.get_total_sol_volume(genesisTxn)
        if hasattr(self, 'initial_sol_amt'):
            return self.initial_sol_amt
        genesis_txns = self.get_genesis_signature(pair)
        self.initial_sol_amt = self.get_total_sol_volume(genesis_txns)
        return self.initial_sol_amt

    def get_initial_token_amt(self, pair=None):
        genesisTxn = self.get_genesis_txn(pair)
        if pair:
            return self.get_total_token_volume(genesisTxn)
        if hasattr(self, 'initial_token_amt'):
            return self.initial_token_amt
        genesis_txns = s
        elf.get_genesis_signature(pair)
        self.initial_token_amt = self.get_total_token_volume(genesis_txns)
        return self.initial_token_amt

    def get_txns(self, txns=None):
        if not txns:
            if hasattr(self, 'all_txns'):
                return self.all_txns
        return if_signature_get_txn(txns)

    def get_pool_creator(self, pair=None):
        # Fixing typo: use 'user_address'
        if pair:
            return pair.get('user_address')
        if hasattr(self, 'owner_address'):
            return self.owner_address
        self.owner_address = self.pair.get('user_address')
        return self.owner_address

    def get_genesis_signature(self, pair=None):
        if pair:
            return pair.get('signature')
        if hasattr(self, 'genesis_signature'):
            return self.genesis_signature
        self.genesis_signature = self.pair.get('signature')
        return self.genesis_signature

    def get_pair_creation_time(self, pair=None):
        if pair:
            return pair.get('creation_time')
        if hasattr(self, 'creation_time'):
            return self.creation_time
        self.creation_time = self.pair.get('creation_time')
        return self.creation_time

    def get_all_pair_txns(self, pair=None):
        pair = pair or self.pair
        pair_txns = get_txns_for_pair_from_pair_id(pair)
        unique_txns = self.remove_duplicates(pair_txns)
        sorted_txns = self.sort_txns_by_time(unique_txns)
        return sorted_txns

    def clean_txns(self, txns):
        """Clean, sort, and enumerate transactions."""
        unique_txns = self.remove_duplicates(txns)
        sorted_txns = self.sort_txns_by_time(unique_txns)
        enumerated_txns = self.enumerate_txns(sorted_txns)
        return enumerated_txns

    def remove_duplicates(self, transactions):
        """Remove duplicates from transactions using a composite key."""
        key_fields = ['invocation_number', 'isbuy', 'log_id', 'mint', 'timestamp', 'user_address', 'signature']
        seen_keys = set()
        unique_transactions = []
        for txn in transactions:
            composite_key = '-'.join(str(txn.get(field)) for field in key_fields)
            if composite_key not in seen_keys:
                seen_keys.add(composite_key)
                unique_transactions.append(txn)
        return unique_transactions

    def sort_txns_by_time(self, txns):
        """Sort transactions by their timestamp."""
        return sorted(txns, key=lambda x: x.get('timestamp', 0))

    def enumerate_txns(self, txns):
        """Assign an enumeration to each transaction."""
        for i, txn in enumerate(txns):
            txn['txn_number'] = i
        return txns

    def calculate_time_interval(self, txns=None):
        """Calculate the interval between the first and last transaction in minutes."""
        txns = self.get_txns(txns)
        if len(txns) < 2:
            return None
        first_time = txns[0].get('timestamp', 0)
        last_time = txns[-1].get('timestamp', 0)
        return (last_time - first_time) / 60  # minutes

    def get_token_amounts(self, txns=None):
        """Calculate net token amounts for each transaction."""
        txns = self.get_txns(txns)
        return [float(txn.get('token_amountui', 0)) * (1 if get_is_buy(txn) else -1) for txn in txns]

    def get_token_volume_amounts(self, txns=None):
        """Calculate absolute token volume amounts for each transaction."""
        amounts = self.get_token_amounts(txns)
        return [abs(token) for token in amounts]

    def get_total_token_volume(self, txns=None):
        """Return the total volume of token amounts."""
        txns = self.get_txns(txns)
        # Fixed typo: using self.get_token_volume_amounts instead of an undefined variable
        return sum(self.get_token_volume_amounts(txns))

    def get_net_token_amount(self, txns=None):
        """Return the net token amount."""
        txns = self.get_txns(txns)
        return sum(self.get_token_amounts(txns))

    def get_sol_amounts(self, txns=None):
        """Calculate net SOL amounts for each transaction."""
        txns = self.get_txns(txns)
        return [float(get_sol_amount(txn) or 0) * (1 if get_is_buy(txn) else -1) for txn in txns]

    def get_sol_volume_amounts(self, txns=None):
        """Calculate absolute SOL volume amounts for each transaction."""
        amounts = self.get_sol_amounts(txns)
        return [abs(sol) for sol in amounts]

    def get_total_sol_volume(self, txns=None):
        """Return the total volume of SOL amounts."""
        txns = self.get_txns(txns)
        return sum(self.get_sol_volume_amounts(txns))

    def get_net_sol_amount(self, txns=None):
        """Return the net SOL amount."""
        txns = self.get_txns(txns)
        return sum(self.get_sol_amounts(txns))

    def separate_genesis_txn(self, pair=None, txns=None):
        """Separate the genesis transaction from all others."""
        txns = self.get_txns(txns)
        pair = pair or self.pair

        genesis_signature = pair.get('signature')
        all_txns = [txn for txn in txns if txn.get('signature') != genesis_signature]
        genesis_txns = [txn for txn in txns if txn.get('signature') == genesis_signature]

        return genesis_txns, self.clean_txns(all_txns)

    def get_sorted_unique_txns(self, signatures):
        """Get unique sorted transactions by a list of signatures."""
        wallet_txns = [self.get_txn_by_signature(sig) for sig in signatures]
        wallet_txns = [txn for txn in wallet_txns if txn]  # Filter None values
        txn_numbers = [txn.get('txn_number', 0) for txn in wallet_txns]
        sol_amounts = [abs(sol) for sol in self.get_sol_amounts()]
        sol_volumes = [abs(sol_amounts[i]) + abs(sol_amounts[i - 1]) if i > 0 else abs(sol_amounts[i])
                       for i in range(len(sol_amounts))]
        keys = ['signature', 'sol_amount', 'volume', 'txn_number']
        combined = [dict(zip(keys, values)) for values in zip(signatures, sol_amounts, sol_volumes, txn_numbers)]
        return combined

    def total_profit(self, profits):
        """Calculate total profit from a dictionary of profits."""
        return sum(values['profits']['sol'] for key, values in profits.items())

    def get_txn_by_signature(self, signature, txns=None):
        """Find and return a transaction by its signature."""
        txns = self.get_txns(txns)
        for txn in txns:
            if txn.get('signature') == signature:
                return txn
        return None
    
    def tally_profits(self,txns=None):
        return tally_profits(txns=txns)

    def get_price(self, virtual_sol_reserves, virtual_token_reserves):
        """Calculate price using reserves (placeholder)."""
        if not virtual_sol_reserves or not virtual_token_reserves:
            return 0
        return virtual_sol_reserves / virtual_token_reserves
